MODULE FileSearchAll;

FROM SYSTEM IMPORT ADR, ADDRESS, BDOS, IORESULT, HLRESULT;
FROM InOut  IMPORT WriteString, WriteLn, Write, WriteInt, WriteHex;
IMPORT Texts;

CONST
  maxDrives = 16;  (* Laufwerke A–P *)
  maxUsers  = 16;  (* Userbereiche 0–15 *)
  CR        = 15C; (* Carriage Return in Oktal *)

TYPE
  CmdLineBuffer = ARRAY [0..127] OF CHAR;

VAR
  CmdLine[0080H] : CmdLineBuffer;
  fcb: ARRAY [0..36] OF CHAR;
  outputFile: TEXT;
  outputEnabled: BOOLEAN;
  dma: ARRAY [0..3], [0..31] OF CHAR;
  searchPattern: ARRAY [0..12] OF CHAR;
  c: CHAR;
  i, j, len: CARDINAL;

PROCEDURE SetDMA(addr: ADDRESS);
BEGIN
  BDOS(26, addr);
END SetDMA;

PROCEDURE SetUser(user: INTEGER);
BEGIN
  BDOS(32, user);
END SetUser;

PROCEDURE SelectDrive(d: INTEGER): BOOLEAN;
BEGIN
  BDOS(14, d);
  RETURN IORESULT # 255;
END SelectDrive;

PROCEDURE PrepareFCB(pattern: ARRAY OF CHAR);
VAR i, dotPos: CARDINAL;
BEGIN
  FOR i := 0 TO 35 DO fcb[i] := 0C; END;

  dotPos := 0;
  WHILE (dotPos < 12) AND (pattern[dotPos] # '.') AND (pattern[dotPos] # 0C) DO
    INC(dotPos);
  END;

  FOR i := 0 TO 7 DO
    IF i < dotPos THEN
      IF (pattern[i] >= 'a') AND (pattern[i] <= 'z') THEN
        fcb[i+1] := CHR(ORD(pattern[i]) - ORD('a') + ORD('A'));
      ELSIF pattern[i] = '*' THEN
        fcb[i+1] := '?';
      ELSE
        fcb[i+1] := pattern[i];
      END;
    ELSE
      fcb[i+1] := '?';
    END;
  END;

  IF pattern[dotPos] = '.' THEN
    FOR i := 0 TO 2 DO
      IF pattern[dotPos + 1 + i] = 0C THEN
        fcb[9 + i] := '?';
      ELSIF pattern[dotPos + 1 + i] = '*' THEN
        fcb[9 + i] := '?';
      ELSIF (pattern[dotPos + 1 + i] >= 'a') AND (pattern[dotPos + 1 + i] <= 'z') THEN
        fcb[9 + i] := CHR(ORD(pattern[dotPos + 1 + i]) - ORD('a') + ORD('A'));
      ELSE
        fcb[9 + i] := pattern[dotPos + 1 + i];
      END;
    END;
  ELSE
    FOR i := 0 TO 2 DO fcb[9 + i] := '?'; END;
  END;
END PrepareFCB;

PROCEDURE PrintDirectoryEntry(entry: ARRAY OF CHAR);
VAR i: INTEGER;
BEGIN
  WriteString('  ');
  FOR i := 1 TO 8 DO
    IF entry[i] # ' ' THEN
      Write(CHR(ORD(entry[i]) MOD 128));
      IF outputEnabled THEN Texts.WriteChar(outputFile, CHR(ORD(entry[i]) MOD 128)); END;
    END;
  END;
  Write('.');
  IF outputEnabled THEN Texts.WriteChar(outputFile, '.'); END;
  FOR i := 9 TO 11 DO
    IF entry[i] # ' ' THEN
      Write(CHR(ORD(entry[i]) MOD 128));
      IF outputEnabled THEN Texts.WriteChar(outputFile, CHR(ORD(entry[i]) MOD 128)); END;
    END;
  END;
  WriteLn;
  IF outputEnabled THEN Texts.WriteChar(outputFile, CR); Texts.WriteChar(outputFile, 12C); END;
END PrintDirectoryEntry;

PROCEDURE ReportDriveError(prefix: ARRAY OF CHAR; driveChar: CHAR);
VAR errorCode: CARDINAL;
BEGIN
  WriteString(prefix); WriteString('Laufwerk '); Write(driveChar);
  errorCode := HLRESULT DIV 256;
  CASE errorCode OF
    1: WriteString(': Disk I/O Error.');
    | 4: WriteString(': Ungültiges Laufwerk.');
    ELSE
      WriteString(': Unbekannter Fehler (H = '); WriteHex(errorCode, 2); WriteString('H)');
  END;
  WriteLn;
END ReportDriveError;

PROCEDURE SearchAndList(pattern: ARRAY OF CHAR; drive: CHAR): BOOLEAN;
BEGIN
  PrepareFCB(pattern);
  SetDMA(ADR(dma));
  BDOS(17, ADR(fcb));
  WHILE IORESULT <= 3 DO
    PrintDirectoryEntry(dma[IORESULT]);
    BDOS(18, ADR(fcb));
  END;
  RETURN (IORESULT = 255) AND ((HLRESULT DIV 256) = 0);
END SearchAndList;

BEGIN
  outputEnabled := FALSE;
  j := 0;
  FOR i := 0 TO 127 DO
    c := CmdLine[i];
    IF (c >= 'a') AND (c <= 'z') THEN c := CHR(ORD(c) - ORD('a') + ORD('A')); END;
    IF (c = '/') AND (CmdLine[i+1] = 'H') THEN
      WriteLn; WriteString('Verwendung: FILESRCH [WILDCARD] [/FILE]'); WriteLn;
      WriteString('Sucht Dateien auf allen Laufwerken und Userbereichen.'); WriteLn;
      WriteString('/FILE aktiviert die Ausgabe in FILELIST.TXT.'); WriteLn;
      RETURN;
    ELSIF (c = '/') AND (CmdLine[i+1] = 'F') THEN
      outputEnabled := TRUE;
    ELSIF (c # ' ') AND (c # 0C) AND (j < 12) THEN
      searchPattern[j] := c; INC(j);
    END;
  END;

  IF j = 0 THEN
    searchPattern := '???????????';
  ELSE
    searchPattern[j] := 0C;
  END;

  IF outputEnabled THEN REWRITE(outputFile, 'FILELIST.TXT'); END;
  BDOS(45, 254);
  WriteString('Suche auf allen Laufwerken und Userbereichen...'); WriteLn;

  FOR i := 0 TO maxDrives - 1 DO
    IF SelectDrive(i) THEN
      FOR j := 0 TO maxUsers - 1 DO
        SetUser(j);
        Write(CHR(ORD('A') + i)); WriteString(': User ');
        WriteInt(j, 2); WriteLn;
        IF NOT SearchAndList(searchPattern, CHR(ORD('A') + i)) THEN
          ReportDriveError('  ', CHR(ORD('A') + i));
        END;
      END;
    ELSE
      ReportDriveError('', CHR(ORD('A') + i));
    END;
  END;

  WriteString('Fertig.'); WriteLn;
  IF outputEnabled THEN CLOSE(outputFile); END;
END FileSearchAll.
